import dataclasses, dill, pathlib


@dataclasses.dataclass(frozen=True)
class FrequencyBand:
    """
    Frequency band object to aid switching to different frequencies and
    periods, and with unique naming of simulations.
    """

    min_frequency_in_hertz: float
    max_frequency_in_hertz: float

    def __post_init__(self):
        """
        Checks to run after creating a frequency band object.
        """
        if self.min_frequency_in_hertz >= self.max_frequency_in_hertz:
            raise ValueError(
                "Mininum frequency needs to be smaller than maximum frequency."
            )

    @property
    def period_band_name(self) -> str:
        """
        Get a string that describes the period band in seconds.
        """
        return (
            f"{1.0/self.max_frequency_in_hertz}_"
            f"{1.0/self.min_frequency_in_hertz}_seconds"
        )

    @property
    def frequency_band_name(self) -> str:
        """
        Get a string that describes the frequency band in hertz.
        """
        return (
            f"{self.min_frequency_in_hertz}_"
            f"{self.max_frequency_in_hertz}_hertz"
        )

    @property
    def min_period_in_seconds(self) -> float:
        """
        Get the minimum period band in seconds, the inverse of the maximum
        frequency in Hertz.
        """
        return 1.0 / self.max_frequency_in_hertz

    @property
    def max_period_in_seconds(self) -> float:
        """
        Get the maximum period band in seconds, the inverse of the minimum
        frequency in Hertz.
        """
        return 1.0 / self.min_frequency_in_hertz

    def save(
        self, filename: pathlib.Path | str, overwrite: bool = False
    ) -> None:
        """
        Save the frequency band object to disk.

        Args:
            filename: The filename to save the frequency band to. Ideally ends
                with extension `.pkl` to indicate a Python object saved using
                the `dill` submodule.
            overwrite: Whether or not to overwrite an existing file.
        """
        if pathlib.Path(filename).exists() and not overwrite:
            raise FileExistsError(f"{filename} already exists.")

        with open(filename, "wb") as file:
            dill.dump(
                (self.min_frequency_in_hertz, self.max_frequency_in_hertz),
                file,
            )

    @classmethod
    def load(cls, filename: pathlib.Path | str) -> "FrequencyBand":
        """
        Load the frequency band object from disk.

        Args:
            filename: The filename to load the frequency band from.
        """
        if not pathlib.Path(filename).exists():
            raise FileNotFoundError(f"{filename} doesn't exist.")

        with open(filename, "rb") as file:
            minf, maxf = dill.load(file)

        return cls(min_frequency_in_hertz=minf, max_frequency_in_hertz=maxf)

    def _repr_html_(self) -> str:
        """
        Frequency band HTML representation to display in Jupyter notebooks.
        """

        prefixes = {
            # fmt: off
            1e-9: "n", 1e-6: "µ", 1e-3: "m",  1: "", 1e3: "k", 1e6: "M", 
            1e9: "G", 1e12: "T",
            # fmt: on
        }

        min_magnitude = max(
            [
                magnitude
                for magnitude in prefixes.keys()
                if self.min_frequency_in_hertz > magnitude
            ]
        )
        max_magnitude = max(
            [
                magnitude
                for magnitude in prefixes.keys()
                if self.max_frequency_in_hertz > magnitude
            ]
        )
        min_freq = self.min_frequency_in_hertz / min_magnitude
        max_freq = self.max_frequency_in_hertz / max_magnitude

        min_freq_str = f"{min_freq:.3g} {prefixes[min_magnitude]}Hz"
        max_freq_str = f"{max_freq:.3g} {prefixes[max_magnitude]}Hz"

        max_period_string = f"{1.0 / self.min_frequency_in_hertz} s"
        min_period_string = f"{1.0 / self.max_frequency_in_hertz} s"

        return (
            f"<div>Frequency band: {min_freq_str} - {max_freq_str}, "
            f"{min_period_string} - {max_period_string}</div>"
        )
